/*
 * Copyright (c) 2003 by Hewlett-Packard Company.  All rights reserved.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE. 
 */

/* Utilities for building data structures that are safe for use	*/
/* from signal handlers in multithreaded applications.		*/

#ifndef HANDLER_SAFE_DATA_H

#define HANDLER_SAFE_DATA_H

#include "atomic_ops.h"

/* Must be called before any other routines in this file.	*/
/* The argument function is called to abort in the event of	*/
/* an error.  If it is NULL, we call abort().			*/
void HSD_malloc_init(void (*abort_fn)(char *msg));

/* Return a pointer to a new memory area of nb bytes.  Abort if	*/
/* this is not possible.					*/
void * HSD_malloc(size_t nb);

/* A simple map data structure that may be accessed from a signal	*/
/* handler.  We map addresses to counter values.  			*/
/* Hsd_malloc_init() must be called before HSD_maps can be used.	*/

#ifndef HSD_MAP_SIZE
#   define HSD_MAP_SIZE 100*1024
#endif

typedef AO_T HSD_map_data_t;

typedef AO_T HSD_map_key_t;

/* Prof_map_entry is private to implementation */
struct HSD_map_entry;
  
typedef struct {
  struct HSD_map_entry * volatile hash[HSD_MAP_SIZE];
} HSD_map;

/* 
 * Return the value of map p at key.  Map values that were not
 * explicitly set or incremented are zero.
 */
HSD_map_data_t
HSD_map_get(HSD_map *p, HSD_map_key_t key);

/* 
 * Increment the value of map p at key by n.  If it was not previously set
 * its value becomes n.  Return the old value.
 */
HSD_map_data_t
HSD_map_add(HSD_map *p, HSD_map_key_t key, long n);

/* 
 * Set the value of map p at key to v.
 */
void
HSD_map_set(HSD_map *p, HSD_map_key_t key, HSD_map_data_t n);

/*
 * Apply a function to all entries inthe map that have ever been touched.
 * The callback a tripe consisting of 
 * (key, data, 3rd arg of hasd_map_apply_all)
 */
void
HSD_map_apply_all(HSD_map *p,
		   void f(HSD_map_key_t, HSD_map_data_t, void *), void *data);


#endif /* HANDLER_SAFE_DATA_H */
