/*
 * Copyright (c) 2003 by Hewlett-Packard Company.  All rights reserved.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE. 
 */ 

#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/mman.h>
#define AO_REQUIRE_CAS	/* Really fetch-and-add	*/
#include "atomic_ops.h"
#include "handler_safe_data.h"

/* A simple malloc implementation that can safely be called from a	*/
/* signal handler.							*/
/* No support for free().						*/

static char * current = 0;

static char * limit;

static void default_abort_fn(char *msg)
{
  write(2, msg, strlen(msg));
  abort();
}

static void (*abort_fn)(char *msg);

#ifdef __LP64__
# define ARENA_SIZE 1024*1024*1024
#else
# define ARENA_SIZE 32*1024*1024
#endif
#define MIN_ARENA_SIZE 1024*1024

void HSD_malloc_init(void (*fn)(char *msg))
{
  size_t size = ARENA_SIZE;
  void * arena;

  while (size > MIN_ARENA_SIZE)
    {
      arena = mmap(0, ARENA_SIZE, PROT_READ | PROT_WRITE,
		   MAP_PRIVATE | MAP_ANONYMOUS, 0, 0);
      if (arena != MAP_FAILED)
	break;
      size >>= 1;
    }
  if (fn == NULL)
    abort_fn = default_abort_fn;
  else
    abort_fn = fn;
  current = (char *)arena;
  limit = current + size;
}

#define ALIGN 16

void * HSD_malloc(size_t nb)
{
  unsigned long result;

  /* Round up to multiple of ALIGN bytes to preserve alignment */
  nb += (ALIGN - 1);
  nb &= ~(ALIGN - 1);
  result = AO_fetch_and_add((volatile AO_T *)(&current), nb);
  if ((char *)(result + nb) > limit)
    abort_fn("HSD_malloc: Out of memory\n");
  return (void *) result;
}

