/*
 * Copyright (c) 2003 by Hewlett-Packard Company.  All rights reserved.
 *
 * This file is covered by the GNU general public license, version 2.
 * see doc/COPYING for details.
 */

#include <stdlib.h>
#include <pthread.h>
#include <stdio.h>
#include "handler_safe_data.h"

typedef void * (* thr_func)(void * thr_func);

typedef int (* test_func)(void);	/* Returns != 0 on success */

void * run_parallel(int nthreads, thr_func f1, test_func t, char *name)
{
  pthread_attr_t attr;
  pthread_t thr[100];
  int i;
  int code;

  fprintf(stderr, "Testing %s\n", name);
  if (nthreads > 100) 
    {
      fprintf(stderr, "run_parallel: requested too many threads\n");
      abort();
    }

  pthread_attr_init(&attr);

  for (i = 0; i < nthreads; ++i)
    {
      if ((code = pthread_create(thr + i, &attr, f1, (void *)(long)i)) != 0)
	{
	  perror("Thread creation failed");
	  fprintf(stderr, "Pthread_create returned %d, thread %d\n", code, i);
	  abort();
        }
    }
  for (i = 0; i < nthreads; ++i)
    {
      if ((code = pthread_join(thr[i], NULL)) != 0)
	{
	  perror("Thread join failed");
	  fprintf(stderr, "Pthread_join returned %d, thread %d\n", code, i);
	  abort();
        }
    }
  if (t())
    {
      fprintf(stderr, "Succeeded\n");
    }
  else
    {
      fprintf(stderr, "Failed\n");
      abort();
    }
  return 0;
}

#define NITERS 1000000

HSD_map map;

void * map_thr(void *id)
{
  int me = (int)(long)id;
  int i;

  for (i = 0; i < NITERS; ++i)
    {
      HSD_map_key_t key = i % 117;
      long incr = i % 17;

      if (me & 1)
	incr = -incr;
      HSD_map_add(&map, key, incr);
    }
  return 0;
}

void map_init()
{
  HSD_map_key_t i;

  for (i = 0; i < 1000; ++i)
    {
      HSD_map_set(&map, i, 17);
    }
}

int map_test(void)
{
  HSD_map_key_t i;
  HSD_map_data_t tmp;

  for (i = 0; i < 1000; ++i)
    {
      if ((tmp = HSD_map_get(&map, i)) != 17)
	{
	  fprintf(stderr, "Incorrect: map[%ld] = %lu\n", i, (unsigned long)tmp);
	  abort(); 
	}
    }
  if ((tmp = HSD_map_get(&map, 1000)) != 0)
    {
      fprintf(stderr, "Map initialization incorrect: map[1000] = %ld\n", 
	      (unsigned long)tmp);
      abort(); 
    }
  return 1;
}
	  
int main()
{
  HSD_malloc_init(NULL);
  map_init();
  run_parallel(6, map_thr, map_test, "Handler-safe-map");
  return 0;
}
